/*
  $Id: JspServlet.java,v 1.51 2000/04/20 07:00:11 alph Exp $ 

  GNUJSP - a free JSP implementation
  Copyright (C) 1999, Yaroslav Faybishenko <yaroslav@cs.berkeley.edu>
  Copyright (C) 1998-1999, Vincent Partington <vinny@klomp.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, =
  USA.
*/

package org.gjt.jsp;

import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.GenericServlet;
import javax.servlet.SingleThreadModel;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpUtils;
import javax.servlet.jsp.HttpJspPage;
import javax.servlet.jsp.JspFactory;
import javax.servlet.jsp.JspWriter; // for jsdk 2.0 compat
import javax.servlet.RequestDispatcher; // for jsdk 2.0 compat
import javax.servlet.ServletRequest; // for jsdk 2.0 compat
import javax.servlet.ServletResponse; // for jsdk 2.0 compat
import java.net.HttpURLConnection; // for jsdk 2.0 compat

import org.gjt.jsp.jsdk20.HttpServletRequestWrapper;
import org.gjt.jsp.jsdk20.HttpServletResponseWrapper;
import org.gjt.jsp.jsdk20.ServletConfigWrapper;
import org.gjt.jsp.jsdk20.ServletContextWrapper;

public class JspServlet extends HttpServlet 
    implements JspMsg
{
    private static final String cvsid = "$Id: JspServlet.java,v 1.51 2000/04/20 07:00:11 alph Exp $";

    public static final String VERSION = "1.0.1";

    // FIXME: should we use that in gnujsp 1.0 in general? Needed? (alph)
    private static final String        lineSeparator
	= System.getProperty("line.separator");

    /** make GNUJSP use packages/sub directories for generated files */ 
    private boolean	usePackages = false;
    private boolean useHostname = false;

	private boolean		keepJava = false; // keep intermediate Java files
    private File		scratchDir;
    private String[]		javacArgs;

	private boolean		debug;
    public boolean getDebug() { return debug; }

	private boolean		negotiation = false;
	private String		defaultLanguage;

    /** this is for extensive path debugging */
    private boolean             pathDebug = false;
    public boolean getPathDebug() { return pathDebug; }

    private File		pageBase;
    private String              denyURI;
    private boolean             checkDeps;
    private boolean             checkClass;
    private Hashtable		pages;
    private JspParser           parser;
    private String		dirSep;
	private String		charset;
        
    // the server environment needs page base and has to use it.
    // This is used when checking initargs for pagebase.
	private boolean jserv = false;
    private boolean usePageBase = false;
    private static boolean hasBrokenPathTranslated = false;
    private static boolean hasBrokenRequestURI = false;
    private static boolean hasRelativeRealPath = false;
    private static boolean ignorePathInfo = false;
    private int time = 0;
 
    static {
	JspFactory.setDefaultFactory (new JspFactoryImpl ());
	Emitter.setEmitter("java", new JavaEmitter());
    }

    public JspServlet () {
	pages = new Hashtable();
    }
    
    /** 
     * This is part of GenericServlet in JSDK 2.1, but for backward
     * compatibility we override.
     */
    public void log(String message, Throwable throwable) {
	if (PageContextImpl.JSDK20) {
	    log(message);
	    log(getStackTrace(throwable));
	} else super.log(message, throwable);
    }

    public void init (ServletConfig config) throws ServletException {
	// Wrap the ServletConfig if necessary; it will be passed to 
	// all compiled pages.
	if (PageContextImpl.JSDK20) config = new ServletConfigWrapper(config, this);
	super.init (config);

	// engine work arounds
	if(PageContextImpl.JSDK20) {
	    String serverInfo = getServletContext().getServerInfo();
	    if("Rhapsody".equals(System.getProperty( "os.name" ))) {
		// always use page base
		usePageBase = true;
	    } else if(serverInfo.startsWith("ApacheJServ/1.")) {
		jserv = true;
	    } else if(serverInfo.startsWith("vqServer/1.9.")) {
		usePageBase = false;
		hasBrokenPathTranslated = true;
		hasBrokenRequestURI = true;
		hasRelativeRealPath = true;
	    } else if(serverInfo.startsWith("Lotus-Domino/5.0")) {
		ignorePathInfo = true;
		hasBrokenPathTranslated = true;
	    }
	}

	if ((jserv || config.getInitParameter ("jserv") != null)
	    && time > 1)
	    return; /* workaround for a really weird bug w/apache jserv*/
	/* init() gets called twice before any request is
	   received, and then once for the first request,
	   but this time without any config args set!
	   I have a suspicition that this is because every time
	   it sees a reference to a servlet in the config (zone)
	   file, it instantiates one.  - Aug. 15, 1999
	*/

	time++;

	String scratchPath     = config.getInitParameter ("scratchdir");
	String compilerCommand = config.getInitParameter ("compiler");

	String pageBaseProp    = config.getInitParameter ("pagebase");    

	checkDeps = !"false".equals(config.getInitParameter ("checkdependencies"));
	checkClass = "true".equals(config.getInitParameter ("checkclass"));
	usePackages = "true".equals(config.getInitParameter ("usepackages"));
	useHostname = "true".equals(config.getInitParameter ("usehostname"));
	dirSep = usePackages ? "._" : "__";
	denyURI = config.getInitParameter("denyuri");
	debug = "true".equals(config.getInitParameter("debug"));
	/** for extensive path debugging */
	pathDebug = "true".equals(config.getInitParameter("pathdebug"));

	keepJava = "true".equals(config.getInitParameter("keepJava"));

	negotiation = "true".equals(config.getInitParameter ("negotiation"));
	defaultLanguage = config.getInitParameter ("defNegLang");

	// eg. DE or US
	String countryProp     = config.getInitParameter ("country"); 
	// eg. de or en
	String languageProp    = config.getInitParameter ("language");

	if(countryProp != null && languageProp != null) {
	    JspConfig.setLocale(new Locale(languageProp, countryProp));
	}

	charset = config.getInitParameter ("charset");

	if (scratchPath == null) {
	    throw new ServletException 
		(JspConfig.getLocalizedMsg(ERR_gnujsp_missing_init_parameter)
		 + ": \"scratchdir\", " + time);
	}

	if (pageBaseProp == null && usePageBase) {
	    throw new ServletException 
		(JspConfig.getLocalizedMsg(ERR_gnujsp_missing_init_parameter)
		+ ": \"pageBase\", " + time);
	}

	if(debug) { 
	    log("Debugging enabled."); 
	    log(cvsid); 
	}

	scratchDir = new File (scratchPath);

	if (compilerCommand == null)
	    compilerCommand = "builtin-javac -classpath %classpath%" +
		File.pathSeparator +
		"%scratchdir% -d %scratchdir% -deprecation -encoding %encoding% %source%";

	StringQuotesTokenizer st = new StringQuotesTokenizer (compilerCommand);

	javacArgs  = new String [st.countTokens ()];

	for (int i = 0;  i < javacArgs.length;  ++i)
	    javacArgs[i] = st.nextToken ();
   
	if (pageBaseProp != null)
	    pageBase = new File (pageBaseProp);

	if(debug) {
	    log("compiler     = "+ qq(compilerCommand));
	    log("scratchdir   = "+ qq(scratchDir.getAbsolutePath()));
	    log("checkDeps    = "+ checkDeps);
	    log("pagebase     = "+ qq(pageBaseProp));
	    log("needsPageBase= "+ usePageBase);
	    log("usePackages  = "+ usePackages);
	}

	if((! scratchDir.exists() && !scratchDir.mkdirs ())
	   || !scratchDir.isDirectory()
	   ) {
	    log(JspConfig.getLocalizedMsg(ERR_gnujsp_could_not_create)
		+" "+qq(scratchDir.getAbsolutePath()));
	    throw new ServletException
		(JspConfig.getLocalizedMsg(ERR_gnujsp_could_not_create)
		 + " scratchDir");
	}
	if(!scratchDir.canWrite()) {
	    log(JspConfig.getLocalizedMsg(ERR_gnujsp_cannot_write_to)
		+": "+qq(scratchDir.getAbsolutePath()));
	    throw new ServletException
		(JspConfig.getLocalizedMsg(ERR_gnujsp_cannot_write_to)
		 + " scratchDir");
	}
	if (PageContextImpl.JSDK20) {  
	    // FIXME: add finally here to remove file? (alph)
	    try { // detect if the file system is case sensitive

		File mixed = new File(scratchDir, "DeleteMe");
		File upper = new File(scratchDir, "DELETEME");

		if(mixed.exists())
		    mixed.delete();
	    
		FileOutputStream fos = new FileOutputStream(mixed);
		fos.close();
	    
		JspParserContext.setCaseInsensitiveFilenames(upper.exists());
		mixed.delete();
		
		mixed=null;
		upper=null;
	    } catch (Exception e) {
		// FIXME let's be more specific here.
		log("file system check exception: " + e);
	    }
	}

	parser = new JspParser(config, this);
	// for use of pagebase for some engines
	parser.setPageBase(pageBaseProp);
	parser.setPathHints(hasRelativeRealPath);
	
	// info may be useful for other debugs, too,
	// but useful for path debugging.
	if(pathDebug) {
	    log("pathdebug.getServInfo: " 
		+ qq(getServletContext().getServerInfo()));
	    log("pathdebug.os.name    : " 
		+ qq(System.getProperty( "os.name" )));
	    log("pathdebug.os.arch    : " 
		+ qq(System.getProperty( "os.arch" )));
	    log("pathdebug.os.version : " 
		+ qq(System.getProperty( "os.version" )));
	    log("pathdebug.java.vendor: "
		+ qq(System.getProperty( "java.vendor" )));
	    log("pathdebug.java.versi : "
		+ qq(System.getProperty( "java.version" )));
	    log("pathdebug.java.clvers: "
		+ qq(System.getProperty( "java.class.version" )));
	}

	if(debug) {
	    log("JspServlet: initialized");
	}
    }

    /**
     * helper for string printing
     */
    private static String qq(String s) {
	if(s == null) {
	    return "null";
	} else {
	    return "'"+s+"'";
	}
    }

    /**
     * Translate a string containing URL-encoded characters, where a '+'
     * represents a space character and %xx represents the equivalent
     * hexadecimal character.
     */
    private static String urlDecode(String val) {
        StringBuffer	buf = new StringBuffer(val.length());
        char		c;

	for (int i = 0; i < val.length(); i++) {
	    c = val.charAt(i);
	    if(c == '%') {
		try {
		    buf.append((char)Integer.parseInt (val.substring (i+1,i+3),
						       16));
		    i += 2;
		    continue;
		} catch(Exception e) { }
	    } else if(c == '+') {
		buf.append(' ');
		continue;
	    }
	    buf.append(c);
	}
        return buf.toString();
    }

    /**
     * Returns the URI of the requested JSP page, which is hidden by
     * servlet containers in various ways.
     *
     * The following are checked, in descending preference:
     * 1. request.getPathInfo() (if non-null and non-empty)
     * 2. request.getRequestURI() (if non-null and implementation works)
     * 3. request.getServletPath() (if non-null)
     * If all of these fail, the method returns null.
     *
     * @param req the request being serviced
     * @returns the URI as a String, or null if it cannot be computed. 
     */
    /*
     * FIXME: I don't see any reason for using getPathInfo()
     * first. From reading the 2.2 Spec I think we should 
     * (as Noel implemented) 
     * 1. request.getServletPath() (need to mark it as broken for jserv!)
     * 2. request.getRequestURI()
     * 3. request.getPathInfo() 
     * (alph)
     */
    private static String requestToJspURI (HttpServletRequest req) {
	// I like java, it's so C-like ... :-) (alph)
	String s;
	return ((!ignorePathInfo 
		 && ((s = req.getPathInfo()) != null)
		 && !"".equals(s)) ? s :
		((!hasBrokenRequestURI && (s = req.getRequestURI()) != null) 
		 ? urlDecode(s) :
		 (((s = req.getServletPath()) != null) ? s :
		  null)));
    }

    /**
     * Finds a Page object for a given URI.  If one does not exist,
     * it is created and added to the hashtable.  The internal Page
     * object is just a placeholder; the JSP page is not parsed or
     * compiled until the Page's process() method is called for the
     * first time.
     * 
     * @param theURI the URI mapping to look up.
     * @returns a Page object which handles requests for the given URI.
     */
    private Page getPageForURI(HttpServletRequest req, String theURI) {
	Page page;
	// suggested by Van Marshall for
	// correct caching on virtual servers
	String key = theURI;
	if (useHostname) {
            key = req.getServerName() + "_" + req.getServerPort() + "_" 
		+ theURI;
        }
        if (pathDebug)
            log("pathdebug.getPageForURI: uri="+theURI+", key=" + key);
 
	// synchronize needed,
	// multiple JspServlet.service() calls
	// for same page are possible here (alph)
	synchronized(pages) {
	    // Create the page if it doesn't exist
	    if ((page = (Page) pages.get (key)) == null) {
		page = new Page (theURI);
		pages.put (key, page);
	    }
	}
	return page;
    }

    /**
     * Process a request from the outside world.
     */
    protected void service (HttpServletRequest request,
			    HttpServletResponse response)
	throws IOException, ServletException {
	String jspURI = null;
	String contentType = "text/html";
	if (charset != null) {
	    contentType += "; charset=" + charset;
	}

	try {
	    if(debug) {
		log("service started");
	    }
	    if(pathDebug) {
		log("PathInfo    = "+qq(request.getPathInfo()));
		log("RequestURI  = "+qq(request.getRequestURI()));
		log("ServletPath = "+qq(request.getServletPath()));
	    }
	    
	    // Determine the requested URI.
	    jspURI = requestToJspURI (request);

		if( !negotiation) xservice( request, response, jspURI);
		else
		{
			// Language negotiation
			String jspURIneg = jspURI;

			String lg;
			Vector al = null;
			int x = 0;
			for( ;;)
			{
				try
				{
					xservice( request, response, jspURIneg);
					break;

				} catch (FileNotFoundException fnf)
				{
					log( "negotiation: "+qq( jspURIneg)+" not found");
					if( al==null)
					{
						al = new Vector();
						String l = request.getHeader( "accept-language");
						String lo;
						int j, k, m;
						int n = l.length();
						for( int i = 0; i<n; i = j+1)
						{
							if(( j = l.indexOf( ",", i))<0) j = n;
							lo = l.substring( i, j);
							if(( k = lo.indexOf( ";"))>-1) lo = lo.substring( 0, k);
							al.addElement( lo);
							if(( m = lo.indexOf( "-"))>-1) al.addElement( lo.substring( 0, m));
						}
						if( al.size()==0 || !al.contains( defaultLanguage)) al.addElement( defaultLanguage);
					}
					if( x>=al.size()) throw fnf;
					lg = ( String) al.elementAt( x++);
					jspURIneg = jspURI.endsWith( ".jsp")
								? jspURI.substring( 0, jspURI.length()-3)+lg+".jsp"
								: jspURI+"."+lg;
				}
			}
		}
	} catch (FileNotFoundException fnf) {
	    if(debug) {
		log("service: file not found: "+ fnf.getMessage());
	    }
	    response.sendError (HttpServletResponse.SC_NOT_FOUND, "JSP File Not Found");
	} catch (ParseException pe) {
	    response.setContentType (contentType);
	    //response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
	    PrintWriter w = new PrintWriter (response.getWriter ());
	    pe.writeHTMLMessage(w);
	    // Should not hurt and vqserver 1.9.17 needs it. (alph)
	    w.flush();
	    return;
	} catch (CompileException ce) {
	    response.setContentType (contentType);
	    response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
	    PrintWriter w = new PrintWriter (response.getWriter ());
	    w.println("<HTML><HEAD><TITLE>GNUJSP Compiler Exception"
		      +"</TITLE></HEAD><BODY><PRE>");
	    w.println(JspConfig.getLocalizedMsg(ERR_gnujsp_error_compiling_source_file)
		      +" " + ce.getURL());
	    w.println("");
	    // FIXME: htmlEncode? (alph)
	    // FIXME: needed in PageContext,too! (alph)
	    // FIXME: use in other exceptions (alph)
	    w.println(transcribeErrors(ce.getURL(), ce.getMessage()));
	    w.println("</PRE></BODY></HTML>");
	    // Should not hurt and vqserver 1.9.17 needs it. (alph)
	    w.flush();
	} catch(NullPointerException e) {
	    log("GNUJSP: Internal error ", e);
	} catch (Throwable t) {
	    log(JspConfig.getLocalizedMsg(ERR_gnujsp_unhandled_exception), t);

	    // FIXME: There are problems with this and flushing.
	    response.setContentType (contentType);
	    response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
	    PrintWriter w = new PrintWriter (response.getWriter ());
	    w.println ("<body bgcolor=\"#ffffff\"><pre>");
	    w.println (JspConfig.getLocalizedMsg
		       (ERR_gnujsp_exception_while_servicing_request)
		       + " "+ jspURI + ":");
	    w.println ("");
	    // FIXME: HTML-Encode?
	    w.print(getStackTrace (t));
	    w.println ("</pre></body>");
	    // Should not hurt and vqserver 1.9.17 needs it. (alph)
	    w.flush();
	}
    }

    /**
     * Process a request internal to GNUJSP.  The service() method calls
     * this method after determining the actual URI of the JSP page.
     * This needs to be public so jsdk20.RequestDispatcherImpl can use it.
     * The caller is responsible for catching any exceptions that may
     * be thrown.
     */
    public void xservice (HttpServletRequest request,
			  HttpServletResponse response,
			  String jspURI)
	throws IOException, ServletException
    {
	// Deny requests beginning with denyURI, if specified.
	if ((denyURI != null) && (jspURI.startsWith(denyURI))) {
	    response.sendError(HttpServletResponse.SC_UNAUTHORIZED);
	}

	// Find or create a Page object to store the details of this URI.
	Page page = getPageForURI(request, jspURI);
	page.process (request, response);
    }

    /** Destroy all servlets generated by this JspServlet. */
    public void destroy() {
	Enumeration e = pages.elements ();
	while (e.hasMoreElements()) {
	    Page p = (Page) e.nextElement();
	    if(p != null) p.destroy (); // leave the null check, just in case (NjB)
	}
	pages.clear();
	super.destroy();
    }

    /**
     * Developers: Change this to ensure that pages get recompiled when
     * you add new code to the system.  Format is YYYYMMDD[SERIAL].
     */
    static long getCompilerVersion() {
	return 1999111600L;
    }

    /** Identifies the version of the GNUJSP servlet. */
    public String getServletInfo() {
	return "GNUJSP " + VERSION; 
    }
    
    /**
     * Transcribe error messages to point to line number
     * in jsp file.
     */
    private String transcribeErrors(URL page, String errors) {
	LineNumberReader	in;
	BufferedReader		in2;
	Vector			map;
	Enumeration		e;
	MapEntry		entry;
	StringBuffer		errBuf;
	String			s, jspFile;
	int			i, j, k, l, javaLineNr, jspLineNr;
	String                  posCodePrefix = Pos.getCodePrefix();
	File                  javaFile = new File(page.getFile());
	String               filePath = javaFile.getPath();

	try {
	    /* Build mapping from java line numbers to jsp file/line# 
	     * combinations.
	     * We could have done this while writing the java file, but using
	     * LineNumberReader is easier, and it would take extra time when
	     * no errors occur.
	     */
	    in = new LineNumberReader(new FileReader(javaFile));
	    in.setLineNumber(1);
	    map = new Vector();
	    try {
		while((s = in.readLine()) != null) {
		    if(s.startsWith(posCodePrefix)) {
			try {
			    MapEntry me = new MapEntry(in.getLineNumber(),
							new Pos(s));
			    map.addElement(me);
			} catch(NumberFormatException nfexc) { 
			} catch(IllegalArgumentException ee) { }
		    }
		}
	    } finally {
		in.close();
	    }

	    /* Now we read every line of the error messages and translate any
	     * file references there.
	     */
	    in2 = new BufferedReader(new StringReader(errors));
	    errBuf = new StringBuffer();
	    try {
		while((s = in2.readLine()) != null) {
		    i = s.indexOf(filePath);
		    if(i != -1) {
			j = i + filePath.length();
			if(j < s.length()-1 
			   && s.charAt(j) == ':' 
			   && Character.isDigit(s.charAt(j+1))) {
			    j++;
			    k = j;
			    while(k < s.length() 
				  && Character.isDigit(s.charAt(k))) {
				k++;
			    }
			    l = k;
			    while(l+1 < s.length() 
				  && s.charAt(l) == ':' 
				  && Character.isDigit(s.charAt(l+1))) {
				l += 2;
				while(l < s.length() 
				      && Character.isDigit(s.charAt(l))) {
				    l++;
				}
			    }
			    
			    try {
				javaLineNr = Integer.parseInt(s.substring(j, k));
				jspFile = null;
				jspLineNr = 0;
				for(e = map.elements(); 
				    e.hasMoreElements(); ) {
				    entry = (MapEntry) e.nextElement();
				    if(entry.javaLineNr > javaLineNr) {
					break;
				    }
				    jspFile = entry.jspPos.getFile();
				    if(entry.jspPos.hasExtraInfo() &&
				       "+".equals(entry.jspPos.getExtraInfo())) {
					// one source code line corresponds to multiple
					// java code lines
					jspLineNr = entry.jspPos.getLine();
				    } else {
					// one source code line corresponds to one
					// java code line
					jspLineNr = entry.jspPos.getLine()
					    + (javaLineNr - entry.javaLineNr);
				    }
				}
				// valid translation found: use it
				if(jspFile != null) {
				    errBuf.append(s.substring(0, i));
				    errBuf.append("<B>").append(jspFile).append((char) ':').append(jspLineNr).append("</B>");
				    errBuf.append("<!-- ").append(s.substring(i, l)).append(" -->");
				    errBuf.append(s.substring(l)).append(lineSeparator);
				    continue;
				}
			    } catch(NumberFormatException nfexc2) { }
			}
		    }
		    errBuf.append(s).append(lineSeparator);
		}
		return errBuf.toString();
	    } finally {
		in2.close();
	    }
	} catch(IOException ioexc) {
	    return errors;
	}
    }
	
    public static File getFileName (File scratchdir, String classname, String ext, boolean mkdir) {
        File resp = new File (scratchdir, classname.replace ('.', File.separatorChar) + ext);
	if (mkdir)
	    (new File (resp.getParent ())).mkdirs ();
        return resp;
    }

    /**
     * Dump a Throwable's stack trace into a String.
     */
    private String getStackTrace(Throwable e) {
	StringWriter sw = new StringWriter();
	PrintWriter pw = new PrintWriter(sw);
	e.printStackTrace(pw);
	pw.flush();
	return sw.toString();
    }

    /**
     * Inner struct class used to store line number and position mappings
     * for compilation error transcription.
     */
    private static class MapEntry {
	int	javaLineNr;
	Pos     jspPos;
	
	MapEntry(int javaLineNr, Pos pos) {
	    this.javaLineNr = javaLineNr;
	    this.jspPos = pos;
	}

    }

    /**
     * Inner class where each instance represents a requested JSP URI.
     * A page has a number of states, from worst to best:
     * -- Target JSP file does not exist (throw FileNotFoundException)
     * -- Target JSP file cannot be parsed (throw ParseException)
     * -- Generated java file cannot be compiled (throw CompileException)
     * -- Generated class cannot be instantiated (throw NullPointerException)
     * -- Fully useful HttpJspPage
     */
    class Page {
	// The virtual path of this JSP page
	private String jspURI;
	
	// The actual path of the JSP file
	private URL jspURL;

	// The generated java file
	private File sourceFile = null;

	// The servlet generated for this JSP page
	private HttpJspPage servlet;

	// The exception that occurred when generating the servlet,
	// or null if all is well.
	private Exception generationException;

	// A classloader instance handling this servlet.
	private JspClassLoader classLoader;

	// The set of dependencies.
	// Dependencies consist of the JSP file itself and any files
	// which are included statically using the include directive.
        private URL[] deps;

	/** 
	 * Constructs an empty page object for the given URI. 
	 * It is important that this method executes quickly because
	 * all requests to GNUJSP synchronize while it's happening.
	 * That's why the real work of generating a servlet is done in
	 * the process() method.
	 */
	Page(String jspURI) {
	    this.jspURI = jspURI;
	}

	/** Process a servlet request destined for the 
	    generated JSP servlet. */
	void process (HttpServletRequest request, HttpServletResponse response)
	    throws ServletException, IOException, 
		   ParseException,
		   CompileException
	{
	    // Only allow one thread to compile this at a time
	    synchronized (this) {
		if (needToRecompile (request)) {
		    generationException = null;
		    try {
			if (jspURL == null) {
			    try {
				if ((jspURL = getResourceImpl(jspURI, request)) == null)
				    throw new ServletException
					(JspConfig.getLocalizedMsg
					 (ERR_gnujsp_internal_error)
					 
					 + ": jspURL==null");
			    } catch (MalformedURLException mue) {
				generationException = mue;
				throw mue;
			    }
			}
		    
			if (debug) log("Parsing " + jspURL);
			StackedReader sr 
			    = new StackedReader(new InputStreamReader(jspURL.openStream()), jspURI);
			// Alternative: parse context info in JspParser
			// by parameter ...
			JspParserContext jpctx = new JspParserContext(jspURI, jspURL);
			if(pathDebug)
			    jpctx.setPathDebug(true);
			sr.setContextInfo(jpctx);
			JspNode body = null;

			try {
				body = parser.parse(sr);
			} catch(ParseException e) {
				throw e;
			} finally {
				sr.close();
			}

			// StackedReader knows what files were included.
			String[] predeps = sr.getAllIncludes();
			deps = new URL [predeps.length];
			StringBuffer concatDeps = new StringBuffer();
			for (int i = 0; i < predeps.length; i++) {
			    if (debug) log("Depends on " + predeps[i]);
			    if (i > 0) concatDeps.append(',');
			    concatDeps.append(predeps[i]);
			    deps[i] = getResourceImpl(predeps[i], request);
			}
			body.setAttribute("gnujspDeps", concatDeps.toString());
			body.setAttribute("gnujspCompilerVersion", 
					  String.valueOf(getCompilerVersion()));
			body.setAttribute("gnujspTimestamp", String.valueOf(getMaxTimestamp()));
			String className = getClassName(jspURI, request);
			String classPart = className;
			String packageName = null;
			int lastDot = className.lastIndexOf ('.');
			if (lastDot != -1) {
			    packageName = className.substring (0, lastDot);
			    classPart = className.substring (lastDot + 1);
			}

			// This is a kludge for JavaEmitter.
			body.setAttribute("package", packageName);
			body.setAttribute("class", classPart);

			sourceFile = JspServlet.getFileName (scratchDir, 
							     className, 
							     ".java", true);
			if (debug) log("Emitting source to " + sourceFile);
			BufferedWriter w = 
			    new BufferedWriter(new FileWriter(sourceFile));
			Emitter.emit(w, body);
			w.close();

			if (debug) log("Compiling " + sourceFile);
			compile(sourceFile, body.getAttribute("charset"));
			if (!keepJava) sourceFile.delete(); // NjB - delete emitted source after successful compile
			// FIXME: Broken if we ever start using packages.
			if (debug) log("Loading " + className);
			try {
			    load(className);
			} catch (Exception e) {
			    // Load failures represent internal errors
			    throw new NullPointerException(e.toString());
			}
			if (debug) log("Initializing servlet " + className);
			// I need a pointer to JspServlet
			// for jsp:include/jsp:forward on jsdk 2.0 (alph)
			// Note that GenericServlet implements ServletConfig
			servlet.init(getServletConfig());
		    } catch (ParseException e) {
			generationException = e;
			throw e;
		    } catch (CompileException e) {
			generationException = e;
			throw e;
		    }
		} // if needed to recompile
	    } // end synchronized

	    // Check for precompilation protocol -- do this w/o calling
	    // getParameter() to preserve request input stream
	    // This is specified in JSP 1.1
	    boolean precompile = false;
	    String query = request.getQueryString();
	    if (query != null) {
		Hashtable h = HttpUtils.parseQueryString(query);
		String[] o = (String[]) h.get("jsp_precompile");
		if (o != null) {
		    // acceptable values are none, "true", "false".
		    if ((o.length == 0) || o[0].equals("") || o[0].equals("true")) 
			precompile = true;
		    else if (!o[0].equals("false")) {
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,  JspConfig.getLocalizedMsg(ERR_sp11_3_4_illegal_jsp_precompile)); // FIXME should provide msg and localize
			return;
		    }
		}
	    }

	    if (!precompile) {
		// Synchronize access to JSPs that implement SingleThreadModel
		// Wes: Why did you wrap the request here? (alph)
		HttpServletRequest useReq = (PageContextImpl.JSDK20) ?
		    new HttpServletRequestWrapper(request) : request;
		if (servlet instanceof SingleThreadModel) {
		    synchronized (servlet) {
			servlet.service(useReq, response);
		    }
		} else {
		    servlet.service(useReq, response); 
		}
	    }
	}
    
	/** 
	 * Returns a URL for the resource specified by path.  If necessary,
	 * relevant information is gathered from the current request.
	 *
	 * In a compliant JSDK 2.1 container, this method simply calls
	 * ServletContext.getResource(path).  In all probability, however,
	 * the container is neither JSDK 2.1 nor compliant, so we have
	 * a grab-bag of workarounds.
	 *
	 * The workaround starts by determining the path to the current
	 * request JSP (which may or may not be the same as the one
	 * we're trying to track down.)
	 *
	 * If a 2.0 container provides a working implementation of 
	 * request.getPathTranslated(), that will give us the full
	 * filename of the requested JSP.  If the path we're asking
	 * for is the same as the request path, we're done.
	 *
	 * If we're asking for the resource to a different path,
	 * it takes more work.
	 *
	 * If getPathTranslated() does not work, we rely on the user to
	 * set the pagebase init parameter, and simply tack the path
	 * onto that.
	 *
	 * If pagebase is not set, we try calling ServletContext.getRealPath
	 * (which is the JSDK 2.0 version of getResource).  In fact this
	 * would probably be the preferred way to do things except for the
	 * fact that Apache JServ steadfastly returns null.
	 *
	 * Some containers (like vqserver) return a relative real path,
	 * so finally we check for that and patch things up accordingly
	 * if necessary.
	 *
	 * @param path the virtual path (URI) to retrieve the resource for
	 * @param request the current request being serviced
	 * @returns a URL, usually of the form file:/path/on/fs/doc.jsp
	 * @exception MalformedURLException from the underlying 
	 *   URL constructor or getResource() call.
	 */
	URL getResourceImpl(String path, HttpServletRequest request) 
			throws MalformedURLException, FileNotFoundException {
		URL retval = null;
		ServletContext sc = getServletContext();

		if (PageContextImpl.JSDK20) {
			// we need full information to decide
			// how to calculate path 
			if(pathDebug) {
				log("pathdebug.jsdk20     : yes");
				log("pathdebug.pagebase   : "+pageBase);
				log("pathdebug.path       : "+qq(path));
				log("pathdebug.getRealPath: "+qq(sc.getRealPath(path)));
				log("pathdebug.getPathTran: "
					  + qq(request.getPathTranslated()));
				log("pathdebug.getReqURI  : "
					  + qq(request.getRequestURI()));
				log("pathdebug.getPathInfo: "+qq(request.getPathInfo()));
				log("pathdebug.getServPath: "
					  + qq(request.getServletPath()));
			}

			// In JSDK 2.0, we need to find the "document root"
			// because there is no getResource() method.

			String fullPath = null;

			if (jserv)
			{
				if (!usePageBase)
				{
					usePageBase = true;
					String pageBaseProp = (String) request.getAttribute("org.apache.jserv.DOCUMENT_ROOT");
					pageBase = new File (pageBaseProp);
					parser.setPageBase(pageBaseProp);
					parser.setPathHints(hasRelativeRealPath);
				}

				fullPath = pageBase.getAbsolutePath() + path;
				if (pathDebug) System.err.println("fullpath: " + fullPath);
			}
			else
			{

				// BUG! WE WANT PATH FOR THE path VARIABLE, NOT
				// FOR THE CURRENT REQUEST.  FIXME.

				if(!hasBrokenPathTranslated) {
					fullPath = request.getPathTranslated();
				}

				if (fullPath == null) {
					if (pageBase != null) {
						fullPath = 
								  (new File(pageBase, path)).getAbsolutePath();
						if (pathDebug) log("getResIm: fullPath with pageBase="
										   + fullPath);
					} else {
						// Try getting the real path from the context
						fullPath = sc.getRealPath(path);
						if(hasRelativeRealPath) {
							// on vqserver getRealPath returns a relative URL
							// see envdb.txt.
							if(fullPath != null) {
								fullPath = 
										  (new File(fullPath)).getAbsolutePath();
							} else {
								fullPath = 
										  (new File(path)).getAbsolutePath();
							}

						}
						if (pathDebug) log("getResIm: fullPath as realPath=" 
										   + fullPath);
					}

					// this works hopefully on jserv/NT
					if (fullPath == null && pageBase != null)
						fullPath = pageBase + path;
				}

				if(fullPath == null) {
					log(JspConfig.getLocalizedMsg
						  (ERR_gnujsp_could_not_find_resourceimpl));
					throw new FileNotFoundException
							(JspConfig.getLocalizedMsg
							 (ERR_gnujsp_could_not_find_resourceimpl));
				}
			}

			retval = new URL("file:" + fullPath);
			if(debug) log("getResIm: return "+retval.toString());
		} else {
			retval = sc.getResource(path);
			// This is a workaround for problematic engines 
			// Remind me to file a bug report for Jigsaw. --Wes
			if (retval == null) {
				log(JspConfig.getLocalizedMsg(ERR_gnujsp_broken_getresource));
				retval = new URL("file:" + sc.getRealPath(path));
			}
		}
		return retval;
	}

	boolean needToRecompile (HttpServletRequest request) 
	    throws MalformedURLException, FileNotFoundException, ServletException {

	    if (checkClass && (servlet == null)) {
		// See if class is already compiled and current.
		try {
		    // On success, this sets servlet to a new instance.
		    load(getClassName(jspURI, request));

		    // If GNUJSP generated this servlet, we can do some
		    // nice consistency checking.
		    if (servlet instanceof GnuJspPage) {
			// If servlet was generated with a previous
			// version of GNUJSP, we should regenerate it.
			long compilerVersion = ((GnuJspPage) servlet).
			    _gnujspGetCompilerVersion();
			if (compilerVersion < getCompilerVersion()) {
			    servlet = null;
			    return true;
			} else {
			    String[] predeps = null;
			    predeps = ((GnuJspPage) servlet)._gnujspGetDeps();
			
			    deps = new URL [predeps.length];
			    for (int i = 0; i < predeps.length; i++) {
				if (debug) log("Depends on " + predeps[i]);
				deps[i] = getResourceImpl(predeps[i], request);
			    }
			}
		    } // if GnuJspPage

		    // We should really only do this after checking deps
		    servlet.init(getServletConfig());
		} catch (ClassNotFoundException e) { // thrown by load()
		    // We have to do the compiling.
		} catch (IllegalAccessException e) { // thrown by load()
		    generationException = e;
		    // FIXME
		} catch (InstantiationException e) { // thrown by load()
		    generationException = e;
		    // FIXME
		} catch (ServletException e) { // thrown by servlet.init()
		    generationException = e;
		    throw e;
		} catch (FileNotFoundException e) { // thrown by getResourceImpl
		    generationException = e;
		    throw e;
		}
	    }

	    if (servlet == null) return true;
	    if (deps == null) return true; // First time
	   
	    // Check dependencies means to check the timestamps
	    // of the source JSP file and any statically included files.
	    if (checkDeps && (servlet instanceof GnuJspPage)) {
		long compileTime = ((GnuJspPage) servlet)._gnujspGetTimestamp();
		if (debug) log("compileTime was " + compileTime);
		try {
		    if (getMaxTimestamp() > compileTime) return true;
		} catch (IOException e) {
		    generationException = e;
		    return false;
		}
	    }
	    
	    return false;
	}

	/** 
	 * Look at all dependencies (including source JSP) and find the
	 * maximum value.  This will be set as the timestamp attribute of
	 * the generated servlet.  System.currentTimeMillis() is not used
	 * because the JSP files may exist on a separate server than the
	 * servlet container.
	 */
	long getMaxTimestamp() throws MalformedURLException, IOException {
	    long maxstamp = (servlet == null) ? 0 : ((GnuJspPage) servlet)._gnujspGetTimestamp();
	    long l;
	    for (int i = 0;  i < deps.length;  ++i) {
		// Ugly kludge for local files.  Shame on Sun for not
		// providing getLastModified() for FileURLConnections.
		if (deps[i].toString().startsWith("file:")) {
		    l = new File(deps[i].toString().substring(5)).lastModified();
		} else {
		    l = deps[i].openConnection().getLastModified();
		}
		if (debug) log(deps[i] + " timestamp " + l);
		if (l > maxstamp) maxstamp = l;
	    }
	    return maxstamp;
	}

	// Invoke the target language compiler
	void compile (File javaFile, String charset) throws CompileException {
	    String[] compilerArgs;
	    boolean useBuiltinJavac;
	    ByteArrayOutputStream compilerOut = new ByteArrayOutputStream();
	    PrintStream compilerOutStream = new PrintStream(compilerOut, true);
	    int j;
	   
	    // build compiler command line
	    if (javacArgs[0].equals("builtin-javac")) {
		compilerArgs = new String [javacArgs.length - 1];
		useBuiltinJavac = true;
		j = 0;
	    } else {
		compilerArgs = new String [javacArgs.length];
		compilerArgs[0] = javacArgs[0];
		useBuiltinJavac = false;
		j = 1;
	    }
	   
	    compilerOutStream.print (javacArgs[0]);

	    for(int i = 1; i < javacArgs.length; i++) {
		String arg  = javacArgs[i];
		String repl = null;
	   
		if (arg.indexOf ("%classpath%") != -1)
		    repl = System.getProperty ("java.class.path");
		else if (arg.indexOf ("%scratchdir%") != -1)
		    repl = scratchDir.toString ();
		else if (arg.indexOf ("%source%") != -1)
		    repl = javaFile.toString ();
		else if (arg.indexOf ("%encoding%") != -1)
		    repl = (charset != null) 
			? charset : System.getProperty("file.encoding");
	
		int first = arg.indexOf ('%');
		int last  = arg.lastIndexOf ('%');
	
		if (first != -1 && last != -1 && first != last) {
		    compilerArgs[j] = (arg.substring (0, first) + repl +
				       arg.substring (last+1));
		} else {
		    compilerArgs[j] = arg;
		}

		compilerOutStream.print((char) ' ');
		compilerOutStream.print (compilerArgs[j++]);
	    }
	   
	    compilerOutStream.println();
	    if (debug) {
		log("Compiler command: " + compilerOut.toString());
	    }

	    if (useBuiltinJavac) {
			try {
		    Class javacMain = Class.forName("sun.tools.javac.Main");
		    Constructor twoArg = javacMain.getConstructor(new Class[] {
			OutputStream.class, String.class });
		    Object javac = twoArg.newInstance(new Object[] {
			compilerOutStream, "javac" });
		    Method compile = javacMain.getMethod("compile", 
							 new Class[] {
			String[].class });
		    Object compileOut = compile.invoke(javac, new Object[] {
			compilerArgs });
		    if (!((Boolean) compileOut).booleanValue()) {
			generationException = new CompileException(jspURL, compilerOut.toString());
			throw (CompileException) generationException;
		    }
		} catch (Exception e) { // Many types of reflection errors
		    generationException = new CompileException(jspURL, e.getMessage());
		    throw (CompileException) generationException;
		}
	    } else {
		Process		p;
		BufferedReader	stdout = null;
		BufferedReader  stderr = null;
		String		line;
		int		exitValue = -1;
		long		classLastModified;
	
		// FIXME
		//		classLastModified = classFile.lastModified();
	
		try {
		    p = Runtime.getRuntime().exec (compilerArgs);
 		    exitValue = doWaitFor(p, compilerOutStream, compilerOutStream);
 		} catch(InterruptedException ix) {
		} catch(IOException ex) {
		    ex.printStackTrace(compilerOutStream);
		}

		// FIXME
		/*
		  if (classFile.lastModified() == classLastModified) {
		    compilerOutStream.println
			("[no class file has been written]");
		    exitValue = -1;
		}
		*/

		if (exitValue != 0) {
		    throw new CompileException (jspURL, compilerOut.toString ());
		}
	    }
	    //compileTime = System.currentTimeMillis();

	}
	
 /**
  * Workaround for a problem with Process.waitFor() may hang on NT.
  * Process.waitFor() is inconsistent between platforms, which leads to bugs
  * #4098442, #4223650, #4062587, #4103432, and probably others.
  * This problem prevents from using Jikes on NT.
  * This workaround was written from code given in bug #4254231
  */
  private int doWaitFor(Process p, OutputStream _out, OutputStream _err)
  throws InterruptedException
  {
    int exitValue = -1; // returned to caller when p is finished
    InputStream in = p.getInputStream();
    InputStream err = p.getErrorStream();

    boolean finished = false; // Set to true when p is finished
    byte[] buffer = new byte[1024]; // buffer de lecture des sorties du process
    int qtite; // quantite de donnees disponibles

    while(!finished)
    {
      try
      {
        while ((qtite = Math.min(in.available(), buffer.length)) > 0)
        {
          in.read(buffer, 0, qtite);
          if (_out != null)
            _out.write(buffer, 0, qtite);
        }

        while ((qtite = Math.min(err.available(), buffer.length)) > 0)
        {
          err.read(buffer, 0, qtite);
          if (_err != null)
            _err.write(buffer, 0, qtite);
        }

        // Ask the process for its exitValue. If the process
        // is not finished, an IllegalThreadStateException
        // is thrown. If it is finished, we fall through and
        // the variable finished is set to true.
        exitValue = p.exitValue();
        finished = true;
      }
      catch (IOException e)
      {
        e.printStackTrace();
        finished = true;
      }
      catch (IllegalThreadStateException e)
      {
        // Process is not finished yet; Sleep a little to save on CPU cycles
        Thread.sleep(100);
      }
    }

    // return completion status to caller
    return exitValue;
  }

	void load (String className)
	    throws ClassNotFoundException, IllegalAccessException,
		   InstantiationException {
	    classLoader = new JspClassLoader (scratchDir);
	    Class c = classLoader.loadClass (className);
	    servlet = (HttpJspPage) c.newInstance ();
	}

	void destroy () {

	    try {
		servlet.destroy ();

	    } catch (Exception e) {
		/* no code intended - no one needs to know about this */
	    } finally {

		servlet = null;
	    }
	}

	String getClassName (String jspURI, HttpServletRequest request) {
	    StringTokenizer st;
	    StringBuffer    buf;
	    String	    token;
	    char	    c;
	    int		    i;

	    if (useHostname) {
		jspURI = request.getServerName() + "_" + request.getServerPort() + "_" + jspURI;
	    }

	    st  = new StringTokenizer (jspURI, "/");

	    buf = new StringBuffer (jspURI.length() + 32);

	    buf.append("jsp");

	    while (st.hasMoreTokens ()) {
		token = st.nextToken();
	
		buf.append(dirSep);

		for (i = 0;  i < token.length();  i++) {
		    c = token.charAt(i);

		    if (Character.isJavaIdentifierPart (c))
			buf.append (c);
		    else
			buf.append ((char) '_').
			    append (Integer.toHexString (c));
		}
	    }
	
	    return buf.toString();
	}
	/**
	 * we need access to the java source for line number
	 * translation
	 */
	File getSourceFile() {
	    return sourceFile;
	}
    }
}
